/***************************************************************************
 *   Copyright (C) 2004 by Michael Rex                                     *
 *   me@rexi.org                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <sys/file.h>
#include <string.h>
#include <signal.h>

#include "clientmanager.h"
#include "setifiles.h"

#define MAX_CLIENT_ARGS 32
#define PID_FILE "running.pid"


int start_client(const char *boincdir, const char *cmd, const char *args, const char *logfile, const bool appendlog)
{
    char *parsedargs[MAX_CLIENT_ARGS] = {NULL};
    char *argscopy = NULL;
    char *clientcopy = NULL;
    int i = 0;
    int c = 2;
    int return_value = 0;
    pid_t child_pid = 0;
    
    /* starting the client only makes sense if it is not already running */
    if(!is_client_running(boincdir))
    {
        /* 'client' was declared as 'const', this would be discarded if we
           assigned 'client' itself to the array */
        clientcopy = (char *)malloc(strlen(cmd) + 1);
        strncpy(clientcopy, cmd, strlen(clientcopy));

        /* whether we pass arguments to the client program */
        if(args == NULL)
        {
            parsedargs[0] = clientcopy;
            parsedargs[1] = NULL;
        }
        else
        {
            /* first make a copy of the arguments string that we can tamper with */
            argscopy = (char *)malloc(strlen(args) + 1);
            assert(argscopy != NULL);
            strcpy(argscopy, args);

            /* construct the array with the arguments to be passed to the client:
               first argument is command name, second argument is set to the
               beginning of the copied arguments string */
            parsedargs[0] = clientcopy;
            parsedargs[1] = &argscopy[0];

            /* search for the string " -" which we consider denoting the start
               of a new argument */
            for(i = 0; i < (int)strlen(args); i++)
            {
                if(args[i] == ' ' && args[i+1] == '-')
                {
                    argscopy[i] = '\0';
                    parsedargs[c] = &argscopy[i+1];
                    ++c;
                }
            }
            /* execvp requires the last argument to be NULL */
            parsedargs[c] = NULL;
        }

        /* fork off a child process that calls the SETI@home client */
        if((child_pid = fork()) == 0)
        {
            chdir(boincdir);
            if(freopen(logfile, (appendlog ? "a" : "w"), stdout) == NULL)
                fprintf(stderr, "setistat: cannot write to log file: %s\n",
                    strerror(errno));
            if(execvp(cmd, parsedargs) == -1)
                exit(EXIT_FAILURE);
            exit(EXIT_SUCCESS);
        }
        else
        {
            FILE *fd = NULL;
            char *filename = NULL;
            
            filename = (char *)malloc(strlen(boincdir) + strlen(PID_FILE) + 2);
            assert(filename != NULL);
            sprintf(filename, "%s/%s", boincdir, PID_FILE);
            
            if((fd = fopen(filename, "w")) == NULL)
                fprintf(stderr, "setistat: cannot write to %s: %s\n",
                    filename, strerror(errno));
            else
            {
                fprintf(fd, "%d", child_pid);
                fclose(fd);
            }
        }

        free(argscopy);
        return child_pid;
    }
    
    
    //TODO was sinnvolles
    return 0;
}

int stop_client(const char *boincdir)
{
    FILE *fd = NULL;
    char *filename = NULL;
    char line[32] = {'\0'};
    int pid = 0;
    int return_value = 0;
    
    if(is_client_running(boincdir))
    {
        filename = (char *)malloc(strlen(boincdir) + strlen(PID_FILE) + 2);
        assert(filename != NULL);
        sprintf(filename, "%s/%s", boincdir, PID_FILE);
        
        if((fd = fopen(filename, "r")) == NULL)
                fprintf(stderr, "setistat: cannot open %s: %s\n",
                    filename, strerror(errno));
        else
        {
            if(fgets(line, sizeof(line), fd) != NULL)
                pid = atoi(line);
        
            fclose(fd);
            
            unlink(filename);
            
            return_value = kill(pid, SIGTERM);
        }
    }
    
    return return_value;
}

/**
 * Determines whether there is currently a SETI@home client running and
 * processing the work unit given in <code>path</code>.
 * The function tries to get a lock for the lock file in the work unit's
 * directory. If it succeeds we assume that no client is running and processing
 * this work unit.
 * @param path path to the work unit directory
 * @return <code>true</code> if client is running, <code>false</code> if not
 */
bool is_client_running(const char *path)
{
    bool running = false;
    
#ifndef SYSTEM_SUNOS
    char filename[256] = {'\0'};
    int fd = 0;
    
    snprintf(filename, sizeof(filename) - 1, "%s/%s", path, LOCKFILE);
    
    fd = open(filename, O_WRONLY);
    if(lockf(fd, F_TEST, 1) == -1)
    {
        if(errno == EAGAIN || errno == EACCES)
            running = true;
    }
    else
        lockf(fd, F_ULOCK, 1);
        
    close(fd);
#endif

    return running;
}
