/***************************************************************************
 *   Copyright (C) 2004 by Michael Rex                                     *
 *   me@rexi.org                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>

#include "configuration.h"
#include "configparser.h"

#define MAX_LINE_LEN 1024

/* Prototypes */
static void write_default_rc_file(char *rcfile, config_items *cfg);
static int read_line(FILE *fd, char *buffer, int size);
static void parse_line(char *line, config_items *cfg);
static char *strtrim(const char *s);

/**
 *
 */
void parse_rc_file(const char *rcfile, config_items *cfg)
{
    FILE *fd = NULL;
    char *cfgfile = NULL;
    char line[MAX_LINE_LEN] = {'\0'};
    
    /* construct path to rcfile (+2 = 1 '/' + 1 '\0') */
    cfgfile = (char *)malloc(strlen(getenv("HOME")) + strlen(rcfile) + 2);
    sprintf(cfgfile, "%s/%s", getenv("HOME"), rcfile);
    
    write_default_rc_file(cfgfile, cfg);
    
    if((fd = fopen(cfgfile, "r")) == NULL)
        fprintf(stderr, "setistat: cannot open %s: %s\n",
            rcfile, strerror(errno));
    else
    {
        while((read_line(fd, line, sizeof(line))) != -1)
            parse_line(line, cfg);
        
        if(errno != 0)
            fprintf(stderr,
                "setistat: an error occurred while reading %s: %s\n",
                 rcfile, strerror(errno));
                
        if(fclose(fd) != 0)
            fprintf(stderr, "setistat: closing %s failed: %s\n",
                rcfile, strerror(errno));
    }
}

/**
 *
 */
static void write_default_rc_file(char *rcfile, config_items *cfg)
{
    FILE *fd = NULL;
    
    if((fd = fopen(rcfile, "r")) == NULL)
    {
        if((fd = fopen(rcfile, "w")) != NULL)
        {
            fprintf(fd, "%s=%s\n", BOINC_DIR, cfg->boincdir);
//            fprintf(fd, "%s=%s\n", CLIENT_COMMAND, cfg->clientcmd);
            fprintf(fd, "%s=%d\n", REFRESH_RATE, cfg->refresh_rate);
            fprintf(fd, "%s=%d\n", MONOCHROME, cfg->monochrome ? 1 : 0);
            fprintf(fd, "%s=%s\n", LOG_FILE, cfg->logfile);
            fprintf(fd, "%s=%d\n", APPEND_TO_LOG, cfg->appendlog ? 1 : 0);
        
            fclose(fd);
        }
        else
        {
            fprintf(stderr, "setistat: cannot write to %s: %s\n",
                rcfile, strerror(errno));
        }
    }
    else
    {
        fclose(fd);
    }
}

/**
 *
 */
static void parse_line(char *line, config_items *cfg)
{
    char *key = NULL;
    char *value = NULL;
    
    line = strtrim(line);

    /* Skip comments and empty lines */
    if(line[0] == '#' || strcmp(line, "") == 0)
        return;
        
    /* Split the line in key and value */
    key = strtok(line, "=");
    value = strtok(NULL, "=");
    if(value == NULL)
        return;

    /* Parse the line */
    if(strcmp(key, BOINC_DIR) == 0 || strcmp(key, WORKUNIT_DIR) == 0)
    {
        cfg->boincdir = (char *)malloc(strlen(value) + 1);
        assert(cfg->boincdir != NULL);
        strcpy(cfg->boincdir, value);
    }
    else if(strcmp(key, CLIENT_COMMAND) == 0)
    {
        cfg->clientcmd = (char *)malloc(strlen(value) + 1);
        assert(cfg->clientcmd != NULL);
        strcpy(cfg->clientcmd, value);
    }
    else if(strcmp(key, CLIENT_ARGUMENTS) == 0)
    {
        cfg->clientargs = (char *)malloc(strlen(value) + 1);
        assert(cfg->clientargs != NULL);
        strcpy(cfg->clientargs, value);
    }
    else if(strcmp(key, MONOCHROME) == 0)
    {
        cfg->monochrome = (atoi(value) == 1);
    }
    else if(strcmp(key, REFRESH_RATE) == 0)
    {
        cfg->refresh_rate = atoi(value);
    }
    else if(strcmp(key, LOG_FILE) == 0)
    {
        cfg->logfile = (char *)malloc(strlen(value) + 1);
        assert(cfg->logfile != NULL);
        strcpy(cfg->logfile, value);
    }
    else if(strcmp(key, APPEND_TO_LOG) == 0)
    {
        cfg->appendlog = (atoi(value) == 1);
    }
}

/**
 * Reads a line from the given file and stores it in the supplied buffer.
 * Returns the length of the line or -1 when EOF is reached.
 */
static int read_line(FILE *fd, char *buffer, int size)
{
    if(fgets(buffer, size, fd) != NULL)
    {
        /* remove newline-character from string */
        char *c = strchr(buffer, '\n');
        if(c != NULL)
            *c = '\0';
    }

    return (feof(fd) ? -1 : (int)strlen(buffer));
}

/**
 *
 */
static char *strtrim(const char *s)
{
    char *tmp = NULL;
    int start_idx = 0;
    int end_idx = strlen(s) - 1;
    
    tmp = (char *)malloc(strlen(s) + 1);
    assert(tmp != NULL);
    strcpy(tmp, s);

    /* get first non-whitespace character */
    while(isspace(tmp[start_idx]))
        ++start_idx;

    /* get last non-whitespace character */
    while(isspace(tmp[end_idx]))
        --end_idx;

    if(start_idx < end_idx)
    {
        tmp[end_idx + 1] = '\0'; /* cut off trailing whitespace */
        tmp = &tmp[start_idx];   /* cut off leading whitespace */
    }
    else
        tmp[0] = '\0'; /* remove whitespace from empty line */

    return tmp;
}
