/***************************************************************************
 *   Copyright (C) 2004 by Michael Rex                                     *
 *   me@rexi.org                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <ncurses.h>

#include "ncursesui.h"
#include "setiparser.h"
#include "clientmanager.h"
#include "version.h"



/* Color codes */
#define YELLOW_ON_BLUE  1
#define GREEN_ON_BLUE   2
#define WHITE_ON_BLACK  3
#define YELLOW_ON_BLACK 4
#define RED_ON_BLACK    5


/* Offsets */
#define O_CREDIT    1
#define O_STATS     5
#define O_WORKUNITS 10

#define H_HOST      1
#define H_HARDWARE  7
#define H_STATS     14

#define P_PROJECT   1
#define P_USER      5

/* Prototypes */
static void draw_overview_window(void);
static void draw_host_information_window(void);
static void draw_project_window(void);
static WINDOW *get_title_window(const char*title);
static WINDOW *get_status_window(void);
static void update_display(WINDOW *w1, WINDOW *w2, WINDOW *w3);

static void seti_refresh(int signum);
static void zombie_killer(int signum);


/* required global variables */
client_state *state = NULL;
config_items *pc = NULL;

void (*viewed) (void);


/**
 *
 */
void start_ncurses(config_items *cfg)
{
    int ch;
    
    pc = cfg;
    state = get_empty_struct(NULL);
    
    parse_client_state(cfg->boincdir, state);

    /* activate ncurses */        
    if(initscr() == NULL)
    {
        fprintf(stderr, "setistat: error initializing terminal.\n");
        exit(EXIT_FAILURE);
    }
    cbreak();
    noecho();
    keypad(stdscr, true);
    curs_set(0);
    
    /* override user setting if terminal does not support color */
    cfg->monochrome = (cfg->monochrome || !has_colors());
    
    if(!cfg->monochrome)
    {
        start_color();

        init_pair(YELLOW_ON_BLUE,  COLOR_YELLOW, COLOR_BLUE);
        init_pair(GREEN_ON_BLUE,   COLOR_GREEN,  COLOR_BLUE);
        init_pair(WHITE_ON_BLACK,  COLOR_WHITE,  COLOR_BLACK);
        init_pair(YELLOW_ON_BLACK, COLOR_YELLOW, COLOR_BLACK);
        init_pair(RED_ON_BLACK,    COLOR_RED,    COLOR_BLACK);
    }

    /* install signal handlers */
    SIGNAL(SIGALRM, seti_refresh);
    SIGNAL(SIGCHLD, zombie_killer);
    
    alarm(cfg->refresh_rate);


    viewed = &draw_overview_window;
    draw_overview_window();
    
    while((ch = getch()) != 'q')
    {
        switch(ch)
        {
            case 'r':
                viewed();
                break;
            case 's':
                start_client(cfg->boincdir, cfg->clientcmd, cfg->clientargs,
                    cfg->logfile, cfg->appendlog);
                /* give the client a chance to start so it is more likely
                    that the following refresh will find it running */
                sleep(1);
                alarm(0);
                viewed();
                alarm(cfg->refresh_rate);
                break;
            case 'k':
                stop_client(cfg->boincdir);
                sleep(2);
                alarm(0);
                viewed();
                alarm(cfg->refresh_rate);
                break;
            case 'd':
            
                break;
            case '?':
            
                break;
            case 'o':
                draw_overview_window();
                viewed = &draw_overview_window;
                break;
            case 'h':
                draw_host_information_window();
                viewed = &draw_host_information_window;
                break;
            case 'p':
/*                draw_project_window();
                viewed = &draw_project_window;*/
                break;
            case '1':
            
                break;
            case '2':
                break;
            /* ... */
        }
    }
    
    endwin();
}

static void draw_overview_window()
{
    WINDOW *top = NULL;
    WINDOW *center = NULL;
    WINDOW *bottom = NULL;
    int rows = 0;
    int cols = 0;
    double totalcpu = 0.0;
    double progress = 0.0;
    int i = 0;

    /* get console measurements */
    getmaxyx(stdscr, rows, cols);

    /* create windows */
    center = newwin(rows - 2, cols, 1, 0);
    top = get_title_window("Overview");
    bottom = get_status_window();
    
    /* set window attributes depending on terminal's abilities */
    if(!pc->monochrome)
        wbkgd(center, COLOR_PAIR(WHITE_ON_BLACK));

    /* print the captions in the main window */
    if(!pc->monochrome)
        wattron(center, COLOR_PAIR(YELLOW_ON_BLACK) | A_BOLD);
    else
        wattron(center, A_BOLD);

    mvwprintw(center, O_CREDIT, 1, "Credit");
    mvwprintw(center, O_STATS, 1, "Statistics");
    mvwprintw(center, O_WORKUNITS, 1, "Workunits");

    wattrset(center, 0);
    
    /* print credit */
    mvwprintw(center, O_CREDIT + 1, 3, "Total credit: %.2f (%.2f recent average credit)",
        state->user_total_credit, state->user_avg_credit);
    mvwprintw(center, O_CREDIT + 2, 3, "Credit for %s: %.2f (%.2f recent average credit)",
        state->hostname, state->host_total_credit, state->host_avg_credit);
    
    /* print statistics */
    mvwprintw(center, O_STATS + 1, 3, "Workunits: %d (%d active, %d finished, %d queued)",
        state->workunit_count, state->active_count, state->finished_count, 
        state->workunit_count - state->active_count - state->finished_count);
        
    for(i = 0; i < state->workunit_count; ++i)
    {
        workunit_t *wu = &state->workunits[i];
        
        if(wu->state == UPLOADING || wu->state == UPLOADED)
        {
            progress += 1;
            totalcpu += wu->final_cpu;
        }
        else
        {
            progress += wu->progress;
            totalcpu += wu->cputime;
        }
    }
    
    progress = progress / state->workunit_count;
        
    mvwprintw(center, O_STATS + 2, 3, "Completed: %.2f%%",
        progress * 100);
        
    mvwprintw(center, O_STATS + 3, 3, "Total CPU time: %s",
        short_time_string(totalcpu));

        
    /* print workunit overview */    
    for(i = 0; i < state->workunit_count; ++i)
    {
        workunit_t *wu = &state->workunits[i];
        
        switch(wu->state)
        {
            case DOWNLOADING:
            case READY:
                mvwprintw(center, O_WORKUNITS + i + 1, 3, "%d: %s",
                    i + 1, get_state_name(wu->state));
                break;
            case UPLOADING:
            case UPLOADED:
                mvwprintw(center, O_WORKUNITS + i + 1, 3, "%d: %s, 100.00%% (%.2f%% per hour) CPU time: %s",
                    i + 1, get_state_name(wu->state), wu->pph, wu->final_cpu_str);
                break;
            case RUNNING:
                wattron(center, A_BOLD);
                mvwprintw(center, O_WORKUNITS + i + 1, 3, "%d: %s, %.2f%% (%.2f%% per hour) CPU time: %s (left: %s)",
                    i + 1, get_state_name(wu->state), wu->progress * 100,
                    wu->pph, wu->cputime_str,wu->timeleft);
                wattrset(center, 0);
                break;
            default:
                mvwprintw(center, O_WORKUNITS + i + 1, 3, "%d: %s",
                    i + 1, get_state_name(wu->state));
        }
    }

    
    
    



    
    
    update_display(top, center, bottom);
}

static void draw_host_information_window()
{
    WINDOW *top = NULL;
    WINDOW *center = NULL;
    WINDOW *bottom = NULL;
    int rows = 0;
    int cols = 0;
    struct utsname sysinfo;
    
    uname(&sysinfo);

    /* get console measurements */
    getmaxyx(stdscr, rows, cols);

    /* create windows */
    center = newwin(rows - 2, cols, 1, 0);
    top = get_title_window("Host Information");
    bottom = get_status_window();
    
    /* set window attributes depending on terminal's abilities */
    if(!pc->monochrome)
        wbkgd(center, COLOR_PAIR(WHITE_ON_BLACK));

    /* print the captions in the main window */
    if(!pc->monochrome)
        wattron(center, COLOR_PAIR(YELLOW_ON_BLACK) | A_BOLD);
    else
        wattron(center, A_BOLD);

    mvwprintw(center, H_HOST, 1, "Host");
    mvwprintw(center, H_HARDWARE, 1, "Hardware");
    mvwprintw(center, H_STATS, 1, "Stats");

    wattrset(center, 0);
    
    /* print host information */
    mvwprintw(center, H_HOST + 1, 3, "Name: %s  IP address: %s",
        state->hostname, state->ip);
    mvwprintw(center, H_HOST + 2, 3, "Credit: %.2f (%.2f recent average credit)",
        state->host_total_credit, state->host_avg_credit);
    mvwprintw(center, H_HOST + 3, 3, "Operating System: %s %s  Architecture: %s",
        state->os_name, state->os_version, sysinfo.machine);
    mvwprintw(center, H_HOST + 4, 3, "Time zone: UTC %s%d hours",
        state->timezone >= 0 ? "+" : "", state->timezone);
        
    
    mvwprintw(center, H_HARDWARE + 1, 3, "Number of CPUs: %d  CPU type: %s %s",
        state->number_cpus, state->cpu_vendor, state->cpu_model);
    mvwprintw(center, H_HARDWARE + 2, 3, "Floating point speed: %.2f million ops/sec",
        state->float_mips);
    mvwprintw(center, H_HARDWARE + 3, 3, "Integer speed: %.2f million ops/sec",
        state->int_mips);
    mvwprintw(center, H_HARDWARE + 4, 3, "Memory size: %.0f MB",
        state->memory / 1048576);
    mvwprintw(center, H_HARDWARE + 5, 3, "Total disk space: %.2f GB  Free: %.2f GB",
        state->disk_total / 1073741824, state->disk_free / 1073741824);

    mvwprintw(center, H_STATS + 1, 3, "Average upload rate: %.2f KB/sec",
        state->upload / 1024);
    mvwprintw(center, H_STATS + 2, 3, "Average download rate: %.2f KB/sec",
        state->download / 1024);
        
    update_display(top, center, bottom);
}

static void draw_project_window()
{
    WINDOW *top = NULL;
    WINDOW *center = NULL;
    WINDOW *bottom = NULL;
    int rows = 0;
    int cols = 0;

    /* get console measurements */
    getmaxyx(stdscr, rows, cols);

    /* create windows */
    center = newwin(rows - 2, cols, 1, 0);
    top = get_title_window("Project Information");
    bottom = get_status_window();
    
    /* set window attributes depending on terminal's abilities */
    if(!pc->monochrome)
        wbkgd(center, COLOR_PAIR(WHITE_ON_BLACK));

    /* print the captions in the main window */
    if(!pc->monochrome)
        wattron(center, COLOR_PAIR(YELLOW_ON_BLACK) | A_BOLD);
    else
        wattron(center, A_BOLD);

    mvwprintw(center, P_PROJECT, 1, "Project");
    mvwprintw(center, P_USER, 1, "User");

    wattrset(center, 0);
    
    /* print project information */
    mvwprintw(center, P_PROJECT + 1, 3, "Name: %s (%s)",
        state->project_name, state->project_url);
        
    mvwprintw(center, P_USER + 1, 3, "User: %s",
        state->user_name);
    mvwprintw(center, P_USER + 2, 3, "Expected CPU time: %s",
        state->exp_cputime_str);
    
        

    update_display(top, center, bottom);
}

static WINDOW *get_title_window(const char *title)
{
    WINDOW *window = NULL;
    int columns = 0;
    int rows = 0;
    
    getmaxyx(stdscr, rows, columns);
    
    window = newwin(1, columns, 0, 0);

    if(!pc->monochrome)
        wbkgd(window, COLOR_PAIR(YELLOW_ON_BLUE) | A_BOLD);
    else
        wbkgd(window, A_REVERSE | A_BOLD);

    mvwprintw(window, 0, 1, title);

    /* if we can use colors, print the program name in another color */
    if(!pc->monochrome)
        wattron(window, COLOR_PAIR(GREEN_ON_BLUE));

    mvwprintw(window, 0, columns - (strlen(PKG_NAME) + strlen(PKG_VERSION) + 2), "%s %s", PKG_NAME, PKG_VERSION);
    
    return window;
}

static WINDOW *get_status_window()
{
    WINDOW *window = NULL;
    int columns = 0;
    int rows = 0;
    
    getmaxyx(stdscr, rows, columns);
    
    window = newwin(1, columns, rows - 1, 0);

    if(!pc->monochrome)
        wbkgd(window, COLOR_PAIR(YELLOW_ON_BLUE) | A_BOLD);
    else
        wbkgd(window, A_REVERSE | A_BOLD);
    
    mvwprintw(window, 0, 1, "Running: %s", is_client_running(pc->boincdir) ? "yes" : "no");
    mvwprintw(window, 0, columns - (strlen(state->platform) + 1), "%s", state->platform);
    
    
    

    
        
    
    return window;
}

static void update_display(WINDOW *w1, WINDOW *w2, WINDOW *w3)
{
    redrawwin(stdscr);
    refresh();
    wrefresh(w1);
    wrefresh(w2);
    wrefresh(w3);
}

/**
 * This function is called by the signal handler for SIGALRM to update
 * the ncurses display.
 */
static void seti_refresh(int signum)
{
    SIGNAL(SIGALRM, seti_refresh);
    parse_client_state(pc->boincdir, state);
    viewed();
    alarm(pc->refresh_rate);
}

static void zombie_killer(int signum)
{
    wait(NULL);
    SIGNAL(SIGCHLD, zombie_killer);
}

