/***************************************************************************
 *   Copyright (C) 2004 by Michael Rex                                     *
 *   me@rexi.org                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include "setiparser.h"
#include "setifiles.h"
#include "setixml.h"


#define MAX_LINE_LEN 1024
#define MAX_FILENAME_LEN 512
#define BUFF_SIZE 16384



/* Prototypes */
static xmlXPathObjectPtr eval_xpath(xmlDocPtr doc, xmlChar *expr);
static void parse_workunits(xmlDocPtr doc, client_state *p);
static void parse_results(xmlDocPtr doc, client_state *p);
static void parse_active(xmlDocPtr doc, client_state *p);
static void parse_hostinfo(xmlDocPtr doc, client_state *p);
static void parse_project(xmlDocPtr doc, client_state *p);
static void parse_platform(xmlDocPtr doc, client_state *p);
static void parse_netstats(xmlDocPtr doc, client_state *p);

/**
 *
 */
static xmlXPathObjectPtr eval_xpath(xmlDocPtr doc, xmlChar *expr)
{
    xmlXPathContextPtr context;
    xmlXPathObjectPtr result;
    
    context = xmlXPathNewContext(doc);
    result = xmlXPathEvalExpression(expr, context);
    
    if(xmlXPathNodeSetIsEmpty(result->nodesetval))
        result = NULL;
    
    xmlXPathFreeContext(context);

    return result;
}

/**
 *
 */
static void parse_workunits(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlNodeSetPtr nodeset;
    int i = 0;
    
    result = eval_xpath(doc, (xmlChar *)"/client_state/workunit");
    if(result == NULL)
    {
        /* fprintf(stderr, "setistat: no 'workunit' nodes found.\n"); */
        return;
    }
    
    nodeset = result->nodesetval;
    p->workunit_count = nodeset->nodeNr;
    
    for(i = 0; i < nodeset->nodeNr; ++i)
    {
        xmlChar *value = NULL;
        xmlNodePtr node;
        workunit_t wu = { NULL, NULL, 0, 0.0, NULL, false, 0.0, 0.0, -1, 0.0, NULL, 0.0, NULL, NULL, NULL, NULL };

        node = nodeset->nodeTab[i]->xmlChildrenNode;
        
        while(node != NULL)
        {
            if(xmlStrcmp(node->name, (const xmlChar *)WORKUNIT_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.name = (char *)malloc(strlen(value) + 1);
                assert(wu.name != NULL);
                strcpy(wu.name, value);
                
                xmlFree(value);
            }
        
            node = node->next;
        } /* while */
        
        p->workunits[i] = wu;
    } /* for */
}

/**
 *
 */
static void parse_results(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlNodeSetPtr nodeset;
    int i = 0;
    
    result = eval_xpath(doc, (xmlChar *)"/client_state/result");
    if(result == NULL)
    {
        /* fprintf(stderr, "setistat: no 'result' nodes found.\n"); */
        return;
    }
    
    nodeset = result->nodesetval;
    
    for(i = 0; i < nodeset->nodeNr; ++i)
    {
        xmlChar *value = NULL;
        xmlNodePtr node;
        workunit_t wu;
        workunit_t *pwu = NULL;
        int j = 0;        
        char *tmp = NULL;

        node = nodeset->nodeTab[i]->xmlChildrenNode;

        /* get the data we need ... */
                
        while(node != NULL)
        {
            if(xmlStrcmp(node->name, (const xmlChar *)RESULT_WORKUNIT_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.name = (char *)malloc(strlen(value) + 1);
                assert(wu.name != NULL);
                strcpy(wu.name, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)RESULT_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.result_name = (char *)malloc(strlen(value) + 1);
                assert(wu.result_name != NULL);
                strcpy(wu.result_name, value);

                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)RESULT_FINAL_CPU_TIME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.final_cpu = atof(value);

                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)RESULT_STATE) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.state = atoi(value);
                
                if(wu.state == UPLOADING || wu.state == UPLOADED)
                    p->finished_count++;

                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)RESULT_DEADLINE) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.deadline = atof(value);

                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)RESULT_READY) == 0)
            {
                wu.readytoreport = true;
            }
            
            node = node->next;
        } /* while */
        
        /* ... and assign it to the correct workunit */

        for(j = 0; j < p->workunit_count; ++j)
        {
            if(strcmp(wu.name, p->workunits[j].name) == 0)
            {
                pwu = &(p->workunits[j]);
                break;
            }
        }
        
        if(pwu == NULL)
        {
            fprintf(stderr, "setistat: soemthing is strange here... HELP!\n");
            exit(EXIT_FAILURE);
        }
        
        pwu->result_name = (char *)malloc(strlen(wu.result_name) + 1);
        assert(pwu->result_name != NULL);
        strcpy(pwu->result_name, wu.result_name);

        /* final cpu time & string representation */        
        pwu->final_cpu = wu.final_cpu;
        tmp = short_time_string(pwu->final_cpu);
        pwu->final_cpu_str = (char *)malloc(strlen(tmp) + 1);
        assert(pwu->final_cpu_str != NULL);
        strcpy(pwu->final_cpu_str, tmp);
        
        /* percent of workunit processed per hour */
        pwu->pph = (100 / pwu->final_cpu) * 3600;
        
        
        pwu->state = wu.state;
        pwu->deadline = wu.deadline;
        pwu->readytoreport = wu.readytoreport;
    } /* for */
}

/**
 *
 */
static void parse_active(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlNodeSetPtr nodeset;
    int i = 0;    
    
    result = eval_xpath(doc, (xmlChar *)"/client_state/active_task_set/active_task");
    if(result == NULL)
    {
        /* fprintf(stderr, "setistat: no 'active_task' nodes found.\n"); */
        return;
    }
    
    nodeset = result->nodesetval;
    p->active_count = nodeset->nodeNr;

    for(i = 0; i < nodeset->nodeNr; ++i)
    {
        xmlChar *value = NULL;
        xmlNodePtr node;
        workunit_t wu;
        workunit_t *pwu = NULL;
        int j = 0;        
        char *tmp = NULL;
        
        node = nodeset->nodeTab[i]->xmlChildrenNode;
        
        /* get the data we need ... */
        
        while(node != NULL)
        {
            if(xmlStrcmp(node->name, (const xmlChar *)ACTIVE_SLOT) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.slot = atoi(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)ACTIVE_PROGRESS) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.progress = atof(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)ACTIVE_RESULT_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.result_name = (char *)malloc(strlen(value) + 1);
                assert(wu.result_name != NULL);
                strcpy(wu.result_name, value);

                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)ACTIVE_CPU_TIME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                wu.cputime = atof(value);
                
                xmlFree(value);
            }
            
            node = node->next;
        } /* while */
        
        /* ... and assign it to the correct workunit */

        for(j = 0; j < p->workunit_count; ++j)
        {
            if(strcmp(wu.result_name, p->workunits[j].result_name) == 0)
            {
                pwu = &(p->workunits[j]);
                break;
            }
        }
        
        if(pwu == NULL)
        {
            fprintf(stderr, "setistat: something is strange here... HELP!\n");
            exit(EXIT_FAILURE);
        }
        
        pwu->slot = wu.slot;
        pwu->progress = wu.progress;
        pwu->cputime = wu.cputime;
        
        pwu->state = 999;
        
        /* create string representation of CPU time */
        tmp = short_time_string(pwu->cputime);
        pwu->cputime_str = (char *)malloc(strlen(tmp) + 1);
        assert(pwu->cputime_str != NULL);
        strcpy(pwu->cputime_str, tmp);
                
        /* percent of workunit processed per hour */
         
        /* time left till workunit is completed */
        if(pwu->progress == 0)
            tmp = short_time_string(0);
        else
            tmp = short_time_string((pwu->cputime / pwu->progress) - pwu->cputime);
            
        pwu->timeleft = (char *)malloc(strlen(tmp) + 1);
        assert(pwu->timeleft != NULL);
        strcpy(pwu->timeleft, tmp);
        
        /* total time expected to process workunit */
        if(pwu->progress == 0)
            tmp = short_time_string(0);
        else
            tmp = short_time_string(pwu->cputime / pwu->progress);
            
        pwu->timetotal = (char *)malloc(strlen(tmp) + 1);
        assert(pwu->timetotal != NULL);
        strcpy(pwu->timetotal, tmp);
        
        /* percent of workunit processed per hour */
        pwu->pph = (100 / pwu->cputime) * 3600;

    } /* for */
}


/**
 *
 */
static void parse_hostinfo(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlNodeSetPtr nodeset;
    int i = 0;
    
    result = eval_xpath(doc, (xmlChar *)"/client_state/host_info");
    if(result == NULL)
        return;
        
    nodeset = result->nodesetval;

    for(i = 0; i < nodeset->nodeNr; ++i)
    {
        xmlChar *value = NULL;
        xmlNodePtr node;
        
        node = nodeset->nodeTab[i]->xmlChildrenNode;
        
        while(node != NULL)
        {
            if(xmlStrcmp(node->name, (const xmlChar *)DOMAIN_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->hostname = (char *)malloc(strlen(value) + 1);
                assert(p->hostname != NULL);
                strcpy(p->hostname, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)IP_ADDRESS) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->ip = (char *)malloc(strlen(value) + 1);
                assert(p->ip != NULL);
                strcpy(p->ip, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)NUMBER_CPUS) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->number_cpus = atoi(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)CPU_VENDOR) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->cpu_vendor = (char *)malloc(strlen(value) + 1);
                assert(p->cpu_vendor != NULL);
                strcpy(p->cpu_vendor, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)CPU_MODEL) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->cpu_model = (char *)malloc(strlen(value) + 1);
                assert(p->cpu_model != NULL);
                strcpy(p->cpu_model, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)TIMEZONE) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->timezone = atoi(value) / 3600;
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)OS_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->os_name = (char *)malloc(strlen(value) + 1);
                assert(p->os_name != NULL);
                strcpy(p->os_name, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)OS_VERSION) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->os_version = (char *)malloc(strlen(value) + 1);
                assert(p->os_version != NULL);
                strcpy(p->os_version, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)FLOAT_OPS) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->float_mips = atof(value) / 1000000;
            
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)INTEGER_OPS) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->int_mips = atof(value) / 1000000;
            
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)MEMORY_BYTES) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->memory = atof(value);
            
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)DISK_TOTAL_BYTES) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->disk_total = atof(value);
            
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)DISK_FREE_BYTES) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->disk_free = atof(value);
            
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)HOST_CALCULATED) == 0)
            {
                char *tmp = NULL;
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->calculated = atof(value);
                
                tmp = short_time_string(p->calculated);
                p->calculated_str = (char *)malloc(strlen(tmp) + 1);
                assert(p->calculated_str != NULL);
                strcpy(p->calculated_str, tmp);
                
                xmlFree(value);
            }
        
            node = node->next;
        } /* while */
    } /* for */    
}

/**
 *
 */
static void parse_project(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlNodeSetPtr nodeset;
    int i = 0;
    
    result = eval_xpath(doc, (xmlChar *)"/client_state/project");
    if(result == NULL)
        return;
        
    nodeset = result->nodesetval;

    for(i = 0; i < nodeset->nodeNr; ++i)
    {
        xmlChar *value = NULL;
        xmlNodePtr node;
        
        node = nodeset->nodeTab[i]->xmlChildrenNode;
        
        while(node != NULL)
        {
            if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_USER_TOTAL_CREDIT) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->user_total_credit = atof(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_USER_AVG_CREDIT) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->user_avg_credit = atof(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_HOST_TOTAL_CREDIT) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->host_total_credit = atof(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_HOST_AVG_CREDIT) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->host_avg_credit = atof(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_NAME) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->project_name = (char *)malloc(strlen(value) + 1);
                assert(p->project_name != NULL);
                strcpy(p->project_name, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_URL) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->project_url = (char *)malloc(strlen(value) + 1);
                assert(p->project_url != NULL);
                strcpy(p->project_url, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_USER) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->user_name = (char *)malloc(strlen(value) + 1);
                assert(p->user_name != NULL);
                strcpy(p->user_name, value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)PROJECT_EXP_AVG_CPU) == 0)
            {
                char *tmp = NULL;
                
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->exp_cputime = atof(value);
                
                tmp = short_time_string(p->exp_cputime);
                p->exp_cputime_str = (char *)malloc(strlen(tmp) + 1);
                assert(p->exp_cputime_str != NULL);
                strcpy(p->exp_cputime_str, tmp);

                xmlFree(value);
            }
            
            node = node->next;
        } /* while */
    } /* for */    
}

/**
 *
 */
static void parse_platform(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlChar *value = NULL;
    
    result = eval_xpath(doc, (xmlChar *)"/client_state/platform_name");
    if(result == NULL)
        return;
        
    value = xmlNodeListGetString(doc, result->nodesetval->nodeTab[0]->xmlChildrenNode, 1);

    p->platform = (char *)malloc(strlen(value) + 1);
    assert(p->platform != NULL);
    strcpy(p->platform, value);
            
    xmlFree(value);
}

/**
 *
 */
static void parse_netstats(xmlDocPtr doc, client_state *p)
{
    xmlXPathObjectPtr result;
    xmlNodeSetPtr nodeset;
    int i = 0;
    
    result = eval_xpath(doc, (xmlChar *)XPATH_NETSTATS);
    if(result == NULL)
        return;
        
    nodeset = result->nodesetval;

    for(i = 0; i < nodeset->nodeNr; ++i)
    {
        xmlChar *value = NULL;
        xmlNodePtr node;
        
        node = nodeset->nodeTab[i]->xmlChildrenNode;
        
        while(node != NULL)
        {
            if(xmlStrcmp(node->name, (const xmlChar *)NET_UPLOAD) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->upload = atof(value);
                
                xmlFree(value);
            }
            else if(xmlStrcmp(node->name, (const xmlChar *)NET_DOWNLOAD) == 0)
            {
                value = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
                
                p->download = atof(value);
                
                xmlFree(value);
            }
            
            node = node->next;
        } /* while */
    } /* for */    
}



/**
 *
 */
void parse_client_state(const char *path, client_state *p)
{
    char filename[MAX_FILENAME_LEN] = { '\0' };
    xmlDocPtr doc;
    xmlNodePtr node;
    
    /* clean client_state */
    p = get_empty_struct(p);
    
    snprintf(filename, sizeof(filename) - 1, "%s/%s", path, CLIENT_STATE);
    
    /* parse document */
    
    if((doc = xmlParseFile(filename)) == NULL)
    {
        fprintf(stderr, "setistat: failed to parse document\n");
        exit(EXIT_FAILURE);
    }
    
    if((node = xmlDocGetRootElement(doc)) == NULL)
    {
        fprintf(stderr, "setistat: document empty.\n");
        xmlFreeDoc(doc);
        exit(EXIT_FAILURE);
    }
    
    parse_workunits(doc, p);
    parse_results(doc, p);
    parse_active(doc, p);
    parse_hostinfo(doc, p);
    parse_project(doc, p);
    parse_platform(doc, p);
    parse_netstats(doc, p);
}

 


/**
 * Converts the seconds stored by SETI@home to a string of the format
 * 'h:mm:ss'
 * @param sec the seconds to convert
 * @return pointer to a static buffer containing the result
 */
char *short_time_string(double sec)
{
    unsigned int hours = 0;
    unsigned int minutes = 0;
    unsigned int seconds = 0;
    static char buff[256] = {'\0'};

    hours    = sec / 3600;
    sec     -= hours * 3600;
    minutes  = sec / 60;
    sec     -= minutes * 60;
    seconds  = sec;

    snprintf(buff, sizeof(buff) - 1, "%0d:%02d:%02d", hours, minutes, seconds);

    return buff;
}

char *get_state_name(int state)
{
    switch(state)
    {
        case 0:
            return "0 (unknown)";
        case 1:
            return "downloading";
        case 2:
            return "ready";
        case 3:
            return "3 (unknown)";
        case 4:
            return "uploading";
        case 5:
            return "uploaded";
        case 6:
            return "6 (unknown)";
        case 999:
            return "running";
        default:
            return "totally unknown";
    }
}

/**
 * This function returns a struct to holds information extracted from the
 * client.
 * Either a new struct is created or an existing one is cleared.
 * THIS IS THE ONLY PLACE WHERE free(3) IS CALLED FOR ELEMENTS OF THE STRUCT!
 */
client_state *get_empty_struct(client_state *p)
{
    static client_state s;
    

    if(p == NULL)
    {
        s.hostname = NULL;
        s.ip = NULL;
        s.number_cpus = 0;
        s.float_mips = 0.0;
        s.int_mips = 0.0;
        s.cpu_model = NULL;
        s.cpu_vendor = NULL;
        s.os_name = NULL;
        s.os_version = NULL;
        s.timezone = 0;
        s.memory = 0.0;
        s.disk_total = 0.0;
        s.disk_free = 0.0;
        s.calculated = 0.0;
        s.calculated_str = NULL;
        s.project_name = NULL;
        s.project_url = NULL;
        s.user_name = NULL;
        s.user_total_credit = 0.0;
        s.user_avg_credit = 0.0;
        s.host_total_credit = 0.0;
        s.host_avg_credit = 0.0;
        s.exp_cputime = 0.0;
        s.exp_cputime_str = NULL;
        s.workunit_count = 0;
        s.finished_count = 0;
        s.active_count = 0;
        s.upload = 0.0;
        s.download = 0.0;
        s.platform = NULL;
        
        return &s;
    }
    else
    {
        free(p->hostname);
        p->hostname = NULL;
        free(p->ip);
        p->ip = NULL;
        p->number_cpus = 0;
        p->float_mips = 0.0;
        p->int_mips = 0.0;
        free(p->cpu_model);
        p->cpu_model = NULL;
        free(p->cpu_vendor);
        p->cpu_vendor = NULL;
        free(p->os_name);
        p->os_name = NULL;
        free(p->os_version);
        p->os_version = NULL;
        p->timezone = 0; 
        p->memory = 0.0;
        p->disk_total = 0.0;
        p->disk_free = 0.0;
        p->calculated = 0.0;
        free(p->calculated_str);
        p->calculated_str = NULL;
        free(p->project_name);
        p->project_name = NULL;
        free(p->project_url);
        p->project_url = NULL;
        free(p->user_name);
        p->user_name = NULL;
        p->user_total_credit = 0.0;
        p->user_avg_credit = 0.0;
        p->host_total_credit = 0.0;
        p->host_avg_credit = 0.0;
        p->exp_cputime = 0.0;
        free(p->exp_cputime_str);
        p->exp_cputime_str = NULL;
        p->workunit_count = 0;
        p->finished_count = 0;
        p->active_count = 0;
        p->upload = 0.0;
        p->download = 0.0;
        free(p->platform);
        p->platform = NULL;
        
        return p;
    }
}
